/**
 *    '$RCSfile: CacheManager.java,v $'
 *
 *     '$Author: brooks $'
 *       '$Date: 2005/12/16 18:38:19 $'
 *   '$Revision: 1.8 $'
 *
 *  For Details: http://kepler.ecoinformatics.org
 *
 * Copyright (c) 2003 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN
 * IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY
 * OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */
package org.kepler.objectmanager.cache;

import org.kepler.objectmanager.lsid.KeplerLSID;
import org.kepler.objectmanager.lsid.LSIDResolver;
import org.kepler.objectmanager.lsid.LSIDTree;

import java.util.*;
import java.io.*;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.sql.Statement;
import java.sql.ResultSet;

import org.ecoinformatics.util.Config;

/**
 * This class represents a disk cache of CacheObjects.  The cache manages
 * cache objects by calling their lifecycle event handlers and serializing
 * every time a change is made.  Objects in the cache each have a unique
 * lsid.  Once an object is in the cache, it will not be updated unless
 * the lsid changes.
 * 
 * this class uses hsql to keep track of cache entries.  The tables look like 
 * this:
 * cacheContentTable
 * -----------------
 * varchar name //name of the entry
 * varchar lsid //lsid of the entry
 * varchar date //timestamp of the entry
 * varchar file //filename of the serialized entry
 */
public class CacheManager
{
  private static CacheManager cache = null;
  
  private Hashtable objectHash = new Hashtable();
  
  private Vector listeners = new Vector();
  
  protected static String cachePath = Config.getCacheDirPath();
  private static String objectPath = cachePath + "objects";
  private static String tmpPath = cachePath + "tmp";
  private static String CACHETABLENAME = "cacheContentTable";
  
  private LSIDTree lsidTree = LSIDTree.getInstance();
  
  /**
   * Construct a new CacheManager
   */
  protected CacheManager()
    throws CacheException
  {
    File cachePathFile = new File(cachePath);
    File objectPathFile = new File(objectPath);
    objectPathFile.mkdirs();
    cachePathFile.mkdirs();
    try
    { //check to see if the tables exist
      ResultSet rs = CacheUtil.executeSQLQuery("select name, lsid, file from " + CACHETABLENAME);
    }
    catch(SQLException sqle)
    {
      try
      {
        //the tables don't exist create them
        String sql = "create cached table " + CACHETABLENAME + " (name varchar, " +
          "lsid varchar, date varchar, file varchar)";
        CacheUtil.executeSQLCommand(sql);
      }
      catch(SQLException sqle2)
      {
        sqle2.printStackTrace();
        throw new CacheException("Error creating cache tables.");
      }
    }
  }
  
  /**
   * create a new singleton instance of CacheManager
   */
  public static CacheManager getInstance()
    throws CacheException
  {
    if(cache == null)
    {
      cache = new CacheManager();
    }
    return cache;
  }
  
  /**
   * for testing serialization only.  do not use in production classes.  use
   * getInstance instead.
   */
  public static CacheManager getNewInstance()
    throws CacheException
  {
    return new CacheManager();
  }
  
  /**
   * insert a new CacheObject into the cache.
   */
  public synchronized void insertObject(CacheObject co)
    throws CacheException
  {
    //get the critical info
    String name = co.getName();
    String lsid = co.getLSID().toString();
    String date = new Long(System.currentTimeMillis()).toString();
    String filename = co.getLSID().createFilename();
    File saveFile = new File(cachePath, filename);
    //save the entry to the DB
    try
    {
      CacheUtil.executeSQLCommand("commit");
      String sql = "insert into " + CACHETABLENAME + " (name, lsid, date, file) " +
        "values ('" + name + "', '" + lsid + "', '" + date + "', '" + filename + "');";
      CacheUtil.executeSQLCommand(sql);
      objectHash.put(lsid, co);
      File outputFile = new File(objectPath, filename);
      outputFile.createNewFile();
      FileOutputStream fos = new FileOutputStream(outputFile);
      ObjectOutputStream oos = new ObjectOutputStream(fos);
      oos.writeObject(co); //serialize the CacheObject itself
      CacheUtil.executeSQLCommand("commit");
      lsidTree.addLSID(new KeplerLSID(lsid));
      lsidTree.saveTree();
      notifyListeners(co, "add");
    }
    catch(SQLException sqle)
    {
      try
      {
        CacheUtil.executeSQLCommand("rollback");
      }
      catch(Exception e)
      {
        throw new CacheException("Could not roll back the database after " +
          "error " + sqle.getMessage() + ":::" + e.getMessage());
      }
      throw new CacheException("Could not create hsql entry for new CacheObject: " +
        sqle.getMessage());
    }
    catch(Exception e)
    {
      try
      {
        CacheUtil.executeSQLCommand("rollback");
      }
      catch(Exception e2)
      {
        throw new CacheException("Could not roll back the database after " +
          "error " + e.getMessage() + ":::" + e2.getMessage());
      }
      e.printStackTrace();
      throw new CacheException("Error saving CacheObject to cache: " + 
        e.getMessage());
    }
  }
  
  /**
   * remove the CacheObject with the specified lsid and return it.
   * @param lsid
   */
  public void removeObject(KeplerLSID lsid)
    throws CacheException
  {
    try
    {
      CacheObject co = getObject(lsid);
      CacheUtil.executeSQLCommand("commit");
      String sql = "select file from " + CACHETABLENAME + " where lsid='" + lsid.toString() + "'";
      ResultSet rs = CacheUtil.executeSQLQuery(sql);
      rs.first();
      File f = new File(objectPath, rs.getString("file"));
      f.delete();
      sql = "delete from " + CACHETABLENAME + " where lsid='" + lsid.toString() + "'";
      CacheUtil.executeSQLCommand(sql);
      CacheUtil.executeSQLCommand("commit");
      objectHash.remove(lsid.toString());
      notifyListeners(co, "remove");
    }
    catch(Exception e)
    {
      throw new CacheException("Error removing object " + lsid + " from the " +
        "cache: " + e.getMessage());
    }
  }
  
  /**
   * return the CacheObject with the specified lsid
   * @param lsid
   */
  public CacheObject getObject(KeplerLSID lsid)
    throws CacheException
  {
    //first look in the hash table, if it's not there, go to the db
    CacheObject co = (CacheObject)objectHash.get(lsid.toString());
    if(co == null)
    { //the co isn't in the hash, try the db
      try
      {
        String sql = "select name, lsid, file from " + CACHETABLENAME + 
          " where lsid='" + lsid.toString() + "'";
        ResultSet rs = CacheUtil.executeSQLQuery(sql);
        rs.first();
        String name = rs.getString("name");
        String file = rs.getString("file");
        System.out.println("ser file: " + new File(objectPath, file).getAbsolutePath());
        ObjectInputStream ois = new ObjectInputStream(
          new FileInputStream(new File(objectPath, file)));
        //deserialize the CacheObject
        co = (CacheObject)ois.readObject();
        //add the CacheObject to the hashtable for easier access next time
        objectHash.put(lsid.toString(), co);
      }
      catch(SQLException sqle)
      { //if there is no result, we need to use the lsid resolver to get the object
        try
        {
          LSIDResolver resolver = LSIDResolver.instanceOf();
          InputStream is = resolver.resolveID(lsid);
          //write it to a temp file
          String tmpFileName = System.currentTimeMillis() + ".tmp";
          File tmpFile = new File(tmpPath, tmpFileName);
          FileOutputStream tempFileStream = new FileOutputStream(tmpFile);
          CacheUtil.writeInputStreamToOutputStream(is, tempFileStream);
          //try to open it as a karCO
          try
          {
            KARCacheObject karCO = new KARCacheObject(tmpFile);
          }
          catch(CacheException ce)
          {  //if it isn't, create a RawDataCacheObject
            RawDataCacheObject rdco = new RawDataCacheObject(null, lsid);
            rdco.setData(new FileInputStream(tmpFile));
          }
          //delete the temp file
          //lsidTree.addLSID(lsid);
          //lsidTree.saveTree();
          tmpFile.delete();
        }
        catch(org.kepler.objectmanager.IDNotFoundException idnfe)
        {
          throw new CacheException("Could not resolve lsid " + lsid.toString());
        }
        catch(FileNotFoundException fnfe)
        {
          throw new CacheException("Error creating temp file in the cache: " + 
            fnfe.getMessage());
        }
        catch(IOException ioe)
        {
          throw new CacheException("Error reading or writing data to the cache: " +
            ioe.getMessage());
        }
      }
      catch(Exception e)
      {
        e.printStackTrace();
        throw new CacheException("Error getting object " + lsid.toString() + 
          " from cache: " + e.getMessage());
      }
    }
    return co;
  }
  
  /**
   * return an iterator of all of the CacheObjects in the cache
   */
  public Iterator getCacheObjectIterator()
    throws CacheException
  {
    try
    {
      Vector items = new Vector();
      String sql = "select name, lsid, file from " + CACHETABLENAME;
      ResultSet rs = CacheUtil.executeSQLQuery(sql);
      rs.first();
      
      do
      {
        String name = rs.getString("name");
        String file = rs.getString("file");
        String lsid = rs.getString("lsid");
        CacheObject co = (CacheObject)objectHash.get(lsid);
        if(co == null)
        { //deserialize the object and put it in the hash
          File f = new File(objectPath, file);
          FileInputStream fis = new FileInputStream(f);
          ObjectInputStream ois = new ObjectInputStream(fis);
          co = (CacheObject)ois.readObject();
          objectHash.put(lsid, co);
        }
        items.add(co);
      }
      while(rs.next());
      
      return items.iterator();
    }
    catch(Exception e)
    {
      e.printStackTrace();
      throw new CacheException("Error creating CacheObject iterator. "
              + "Try removing the ~/.kepler directory?: " + 
        e.getMessage());
    }
  }
  
  /**
   * return true of the given lsid has an associated object in the cache
   * @param lsid
   */
  public boolean isContained(KeplerLSID lsid)
    throws CacheException
  {
    try
    {
      String sql = "select name from " + CACHETABLENAME + 
          " where lsid='" + lsid.toString() + "'";
      ResultSet rs = CacheUtil.executeSQLQuery(sql);
      if(rs.next())
      {
        return true;
      }
      else
      {
        return false;
      }
    }
    catch(Exception e)
    {
      throw new CacheException("Error determining contents of cache: " + 
        e.getMessage());
    }
  }
  
  /**
   * clear the cache of all contents
   */
  public void clearCache()
    throws SQLException, CacheException
  {
    CacheUtil.cleanUpDir(new File(objectPath));
    
    String sql = "drop table " + CACHETABLENAME;
    CacheUtil.executeSQLCommand(sql);
    
    sql = "create cached table " + CACHETABLENAME + " (name varchar, " +
          "lsid varchar, date varchar, file varchar)";
    CacheUtil.executeSQLCommand(sql);
  }
  
  /**
   * add a CacheListener to listen for cache events
   */
  public void addCacheListener(CacheListener listener)
  {
    listeners.add(listener);
  }
  
  /**
   * notifies any listeners as to an action taking place.
   */
  private void notifyListeners(CacheObject co, String op)
  {
    CacheEvent ce = new CacheEvent(co);
    
    if(op.equals("add"))
    {
      for(int i=0; i<listeners.size(); i++)
      {
        CacheListener cl = (CacheListener)listeners.elementAt(i);
        cl.objectAdded(ce);
      }
      co.objectAdded();
    }
    else if(op.equals("remove"))
    {
      for(int i=0; i<listeners.size(); i++)
      {
        CacheListener cl = (CacheListener)listeners.elementAt(i);
        cl.objectRemoved(ce);
      }
      co.objectRemoved();
    }
    else if(op.equals("purge"))
    {
      for(int i=0; i<listeners.size(); i++)
      {
        CacheListener cl = (CacheListener)listeners.elementAt(i);
        cl.objectPurged(ce);
      }
      co.objectPurged();
    }
    for(int i=0; i<listeners.size(); i++)
    {
      
    }
  }
}
