/* Base class for actors that process collections.
 *
 * Copyright (c) 2005 Natural Diversity Discovery Project.
 * All rights reserved.
 * 
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 * 
 * IN NO EVENT SHALL THE NATURAL DIVERSITY DISCOVERY PROJECT BE LIABLE 
 * TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR 
 * CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF THIS SOFTWARE AND 
 * ITS DOCUMENTATION, EVEN IF THE NATURAL DIVERSITY DISCOVERY PROJECT
 * HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE NATURAL DIVERSITY DISCOVERY PROJECT SPECIFICALLY DISCLAIMS ANY 
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE NATURAL 
 * DIVERSITY DISCOVERY PROJECT HAS NO OBLIGATION TO PROVIDE MAINTENANCE, 
 * SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.

   @ProposedRating Red (tmcphillips@naturaldiversity.org)
   @AcceptedRating Red (tmcphillips@naturaldiversity.org) 
 */

 package org.nddp;
  
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Stack;

import org.nddp.exceptions.CollectionException;
import org.nddp.tokens.ClosingDelimiterToken;
import org.nddp.tokens.CollectionToken;
import org.nddp.tokens.DataToken;
import org.nddp.tokens.DomainObjectToken;
import org.nddp.tokens.ExceptionToken;
import org.nddp.tokens.LoopTerminationToken;
import org.nddp.tokens.MetadataToken;
import org.nddp.tokens.OpeningDelimiterToken;
import org.nddp.tokens.VariableToken;
import org.nddp.util.Parameters;

import ptolemy.actor.IOPort;
import ptolemy.actor.TypedAtomicActor;
import ptolemy.data.Token;
import ptolemy.data.expr.Parameter;
import ptolemy.kernel.CompositeEntity;
import ptolemy.kernel.Port;
import ptolemy.kernel.util.Attribute;
import ptolemy.kernel.util.IllegalActionException;
import ptolemy.kernel.util.InternalErrorException;
import ptolemy.kernel.util.NameDuplicationException;
 
/**
 * Base class for actors that process collections. The class simplifies the 
 * writing of such actors by providing the following services:
 * <ul>
 * <li>Provides a fire method that calls a handler method for each data token 
 * received.
 * <li>Provides implementations of the non-data handler methods that call 
 * corresponding handlers when a new start of a new collection arrives, when
 * the end of a collection arrives, when a metadata token arrives, etc.
 * <li>Creates collection managers for each incoming collection and passes 
 * references to the collection event handlers.
 * <li>Allows subclasses to specify whether to process and whether to forward
 * on the collections that are received by the actor.  The subclass is not
 * informed of arriving tokens and subcollections part of a collection that 
 * the subclass has chosen not to process.
 * <li>Provides a parameter for specifying types of collections and data that 
 * should be processed.  All other collections and data are passed through 
 * transparently.  
 * </ul>
 * 
 *  @author Timothy M. McPhillips
 */

	public class CollectionActor extends TypedAtomicActor 
		implements CollectionIOPort.NondataHandler {

  	/** 
    	 * Constructs an instance of CollectionActor.
   	 * 
   	 * @param container The run-time container of this actor.
 	 * @param name The name of this actor.
    	 */
	public CollectionActor(CompositeEntity container, String name)
        throws NameDuplicationException, IllegalActionException  {

    	    // call the superclass constructor
        	super(container, name);

        	collectionPath =  Parameters.stringParameter(this, 
    	        "collectionPath", "");
        	
     	// define the icon to represent the actor in Vergil
        _attachText("_iconDescription", "<svg>\n" +
                "<rect x=\"0\" y=\"0\" "
                + "width=\"60\" height=\"20\" "
                + "style=\"fill:white\"/>\n" +
                "</svg>\n");   
    }

	///////////////////////////////////////////////////////////////////
	////                         public variables                  ////
	
    	public static final TokenDisposition DISCARD_TOKEN = 
    	    new TokenDisposition();   
    	
    	public static final TokenDisposition FORWARD_TOKEN = 
    	    new TokenDisposition();
    	
    	public static final ExceptionDisposition DISCARD_EXCEPTION =
    	    new ExceptionDisposition();
    	
    	public static final ExceptionDisposition PROPAGATE_EXCEPTION =
    	    new ExceptionDisposition();
    	
    	public static final CollectionDisposition 
    		IGNORE_AND_DISCARD_COLLECTION = new CollectionDisposition(); 
    	
	public static final CollectionDisposition 
		IGNORE_AND_FORWARD_COLLECTION = new CollectionDisposition();
	
    	public static final CollectionDisposition 
    		PROCESS_AND_DISCARD_COLLECTION = new CollectionDisposition();
    	
	public static final CollectionDisposition 
		PROCESS_AND_FORWARD_COLLECTION = new CollectionDisposition();
	
	public Parameter collectionPath;
	
	///////////////////////////////////////////////////////////////////
	////                         public methods                    //// 

    /** 
     * Handles changes in parameter values.
     *  
     * @param attribute The parameter with a new value.
     */
  	public final void attributeChanged(Attribute attribute) 
		throws IllegalActionException {
            
        if (attribute == collectionPath) {
        		_collectionPath = new CollectionPath(collectionPath);

        } else {
        	    
        	    if (attribute instanceof Parameter) {
	        	    _handleParameterChange((Parameter)attribute, 
	        	            ((Parameter)attribute).getToken());
        	    } else {
        	        super.attributeChanged(attribute);
        	    }
        	}   
  	}
  	
    /** 
     * The method called by the director when the actor is fired. The method
     * reads the next data token from the input port, passes the data token to
     * the handleData method, and then passes the data on via the current
     * collection manager if the collection has not been discarded and the 
     * handleData method did not return DO_NOT_SEND_DATA.
     * @throws IllegalActionException
     * @throws 
     */
    public void fire() throws IllegalActionException {
 
        // call superclass method
        super.fire();
        
        // return immediately if actor has no active input ports
       	if (_inputPort == null) {
       	    return;
       	}
        
        // read next data token from input port, passing a value
        // of false to the unwrapData parameter so that DataTokens
        // can be checked for proper nesting in collections
        Token token = _inputPort.get(false);
        
        if (token == null) {
            return;
        }
        
        // the block below is a kludge to allow collection actors to handle
        // data outside of collections.
        // TODO require child class to enable explicitly the processing of 
        // data outside of collections
        if (_collectionManager == null) {  
            	try {                
            	    _handleData(null, token);
                
            } catch (CollectionException exception) {
                    
            		throw new IllegalActionException(
                		this, exception.getMessage());
            }
        		return;
        }
        
        // assert that if token is a CollectionToken then the token's
        // collection is in the current collection
	 	assert !(token instanceof CollectionToken) || 
	 		((CollectionToken)token).collection() == 
	 		    _collectionManager.collection():
		    "Token outside of collection: " +
		    "\n\t token = " + token + "\n\t token's collection = " + 
		    ((CollectionToken)token).collection() + 
		    "\n\t current collection = " + _collectionManager.collection() +
		    "\n\t actor = " + this;
        
	    	// if not discarding this collection and this is the first data token
	 	// for the collection, inform collection manager accordingly
        if (! _collectionManager.discarded() && 
                ! _collectionManager._firstDataReceived()) {
            
            _collectionManager._announceFirstData();
        }
	    
        _unwrappedDataToken = null;
        _unwrappedDomainObject = null;
        TokenDisposition dataDisposition = _defaultDataDisposition;

    		if (! _collectionManager.ignored() && _typeMatchesTarget(token)) {
    
    		    try {
			    if (_unwrappedDataToken != null) {
			        
				dataDisposition = _handleData(_collectionManager, 
				        _unwrappedDataToken);
				
			    } else {
			        
			        dataDisposition = _handleDomainObject(_collectionManager, 
	                _unwrappedDomainObject);
			    }
			    
    		    } catch(CollectionException exception) {
    		        
				processCollectionException(exception);
    		    }
    		}
    		
		// pass the data on via the collection manager if collection is 
		// not discarded and the disposition of the data token is SEND_DATA
		if (! _collectionManager.discarded() && 
		        dataDisposition == FORWARD_TOKEN) {
		    
        		_collectionManager._passData(token);
    		}
    	}   

	/** 
    	 * Handles closing delimiter tokens for collection actors. Passes the
    	 * collection manager for the delimited collection to the 
    	 * handleCollectionEnd method if the collection is being processed.
    	 *  
    	 * @param token The closing delimiter token.
    	 * @return DO_NOT_SEND_TOKEN in all cases. The collection manager will
    	 * send the token along with the rest of the collection later if the 
    	 * collection is not discarded.
	 * @throws CollectionException
    	 */
	public void handleClosingDelimiter(ClosingDelimiterToken token) 
		throws IllegalActionException {

         // ensure that the closing delimiter is for the current collection  
	 	assert token.collection() == _collectionManager.collection():
		    "Closing delimiter for wrong collection: " +
		    "\n\t token = " + token +
		    "\n\t delimiter's collection = " + token.collection() +
		 	"\n\t current collection = " + _collectionManager.collection() + 	    
		    "\n\t actor = " + this;

	 	// call the collection-end handler if this delimited collection is not 
	 	// being ignored
	 	if (! _collectionManager.ignored() && 
 	        _collectionPath.exposeNonDataEvents()) {
			
	 	    try {
                _handleCollectionEnd(_collectionManager);
            } catch (CollectionException exception) {
		       processCollectionException(exception);
            }
		}
	 	
		// close the collection if it is not being discarded
		if (! _collectionManager.discarded()) {
	    		_collectionManager.closeCollection();
		}
		
		// note that we are no longer below the highest collection type that 
		// should be processed if the closing delimiter is of the highest
		// collection being processed
		if (_belowUpperIgnoredTypes && _collectionManager.collection() == 
		    		_highestProcessedCollection) {
		    
		    // set the flag
			_belowUpperIgnoredTypes = false;
			
			// discard reference to highest collection processed
			_highestProcessedCollection = null;
		}
		
		// revert parameters overridden in this collection to previous values
		HashMap oldParameterValues = _collectionManager._oldParameterValues();
		
		for (Iterator i = oldParameterValues.keySet().iterator();
				i.hasNext(); ) {
		    
		    Parameter parameter = (Parameter)i.next();
		    Token previousValue = (Token)oldParameterValues.get(parameter);
		    _handleParameterChange(parameter, previousValue);
 	        _overriddenParametersMap.put(parameter, previousValue);
		}
    		
		// pop the parent collection manager off the stack and reinstate it
		// as the current collection manager
		_collectionManager = ((IncomingCollectionManager)
		        (_collectionManagerStack.pop()));
		
		// note that we are now above the lowest collection type that should 
		// be processed if closing delimiter is of the lowest collection being
		// processed
		if (! _aboveLowerIgnoredTypes && _collectionManager != null 
		        && _collectionManager.collection() == 
		            _lowestProcessedCollection) {
			
		    // set the flag
		    _aboveLowerIgnoredTypes = true;
		    
		    // discard reference to the lowest collection
			_lowestProcessedCollection = null;
		}
	}
   	   	
    public void handleExceptionToken(ExceptionToken token) 
    		throws IllegalActionException {

        // ensure that each BToken arrives between the opening and closing
        // delimiters of its collection  
	 	assert token.collection() == _collectionManager.collection():
		    "Token outside of collection: " +
		    "\n\t token = " + token + 
		    "\n\t token's collection = " + token.collection() +
		 	"\n\t current collection = " + _collectionManager.collection() + 	    
		    "\n\t actor = " + this;
	 	
	 	TokenDisposition exceptionDisposition = FORWARD_TOKEN;
	 	
	 	if (! _collectionManager.ignored() && 
 	        _collectionPath.exposeNonDataEvents()) {
	 	    
	 	    try {
                exceptionDisposition = _handleExceptionToken(
                        _collectionManager, token);
            } catch (CollectionException exception) {
		       	processCollectionException(exception);
            }
	 	}
	 	
        if (exceptionDisposition == FORWARD_TOKEN) {
	 		_collectionManager._storeInputExceptionToken(token);
	 	}	
	} 


	/** 
    	 * Handles metadata tokens for collection actors. Passes the metadata 
    	 * token and the current collection manager to the handleMetadata
    	 * method if the collection is being processed.
    	 *  
    	 * @param token The metadat delimiter token.
    	 * @return DO_NOT_SEND_TOKEN in all cases. The collection manager will
    	 * send the token along with the rest of the collection later if the
    	 * collection is not discarded.
    	 */
   	public final void handleMetadataToken(MetadataToken token) 
   		throws IllegalActionException {

         // ensure that each BToken arrives between the opening and closing
   	    	// delimiters of its collection  
	 	assert token.collection() == _collectionManager.collection():
		    "Token outside of collection: " +
		    "\n\t token = " + token + 
		    "\n\t token's collection = " + token.collection() +
		 	"\n\t current collection = " + _collectionManager.collection() + 	    
		    "\n\t actor = " + this;

	 	// apply parameter this actor if token is a ParameterToken
	 	if (token instanceof VariableToken) {
			_handleVariableToken((VariableToken) token);
	 	}
	 	
	   	// store and forward the token if the collection is ignored or the 
	 	// metadata handler so requests 
		TokenDisposition metadataDisposition = FORWARD_TOKEN;
		
	 	if (! _collectionManager.ignored() && 
 	        _collectionPath.exposeNonDataEvents()) {
 		    
	 	    try {
                metadataDisposition = 
                    _handleMetadata(_collectionManager, token);
            } catch (CollectionException exception) {
		       	processCollectionException(exception);
	        }
	 	}
 		    
		if (metadataDisposition == FORWARD_TOKEN) {
	    		_collectionManager._storeInputMetadata(token);
	    }
	}	

	/** 
    	 * Handles opening delimiter tokens for collection actors. Creates a
    	 * new collection manager to handle each incoming collection signaled
    	 * by an opening delimiter and sets the disposition of the collection
    	 * according to whether the collection falls between the highest and 
    	 * lowest collection types to be processed and according to the return
    	 * value of handleCollectionStart if it does fall in that range. The 
    	 * new collection is discarded automatically if the parent collection
    	 * was discarded.
    	 *  
    	 * @param token The opening delimiter token.
    	 * @return DO_NOT_SEND_TOKEN in all cases. The collection manager will
    	 * send the token along with the rest of the collection later if the 
    	 * collection is not discarded.
    	 */
    public void handleOpeningDelimiter(OpeningDelimiterToken token) 
    		throws IllegalActionException {

        	// note that we are no longer above the lower collection types to 
        // ignore if an opening delimiter 	is received when the current 
        // collection type is the lowest collection type to process
	 	if (_aboveLowerIgnoredTypes && _collectionManager != null && 
 	        _collectionPath.ignoreCollectionsInside() != Object.class &&
 	        _collectionPath.ignoreCollectionsInside().
 	        		isInstance(_collectionManager.collection())) {	
   	    		
	 	    // indicate that we are no longer above the lower collection 
	 	    // types to ignore
	 	    _aboveLowerIgnoredTypes = false;
			
			// keep a reference to the collection above the the ignored 
	 	    // collections so that its closing delimiter can be recognized 
	 	    // later
			_lowestProcessedCollection = _collectionManager.collection();
		}
		
   		// push the current collection manager onto the collection
	 	// manager stack
   		_collectionManagerStack.push(_collectionManager);

   		// make the current collection manager the parent collection manager
   		_parentCollectionManager = _collectionManager;
   		
   		// create a new collection manager to manage the incoming collection
   		// represented by the delimiter
    		_collectionManager = manageIncomingCollection(
	            token.collection(), _parentCollectionManager, 
    		            	_getDefaultOutputPort());

    		// note that we are below the upper collection types to ignore if 
    		// the type of the incoming collection matches the highest 
    		// collection type to process
		if (! _belowUpperIgnoredTypes && 
	        _collectionPath.ignoreCollectionsOutside().
	        		isInstance(_collectionManager.collection())) {
			
		    	// set the flag indicating we are now below the highest 
		    	// collection type for processing
		    	_belowUpperIgnoredTypes = true;
			
		    	// keep a reference to this collection so that its closing 
		    	// delimiter can be recognized later
		    	_highestProcessedCollection = _collectionManager.collection();
		}
    		
		// discard the collection if the parent collection has been discarded
		if (_parentCollectionManager != null && 
		        _parentCollectionManager.discarded()) {
		    
		    	_collectionManager._discard();
		}
		
		// start processing the collection if it is below the upper ignored 
		// types and the lower ignored types
		if (_belowUpperIgnoredTypes && _aboveLowerIgnoredTypes) {

		    // call the collection start handler
		    CollectionDisposition disposition = PROCESS_AND_FORWARD_COLLECTION;

		    if (_collectionPath.exposeNonDataEvents()) {
	            try {
	                disposition = _handleCollectionStart(_collectionManager);
	            } catch (CollectionException exception) {
			       	processCollectionException(exception);
					disposition = IGNORE_AND_FORWARD_COLLECTION;
	            }
		    }
		    
            // ignore the collection if so requested
		    if (disposition == IGNORE_AND_DISCARD_COLLECTION || 
		            disposition == IGNORE_AND_FORWARD_COLLECTION) {
		        
		        _collectionManager._ignore();
		    }
		    
		    // discard the collection if so requested
		    if (disposition == IGNORE_AND_DISCARD_COLLECTION || 
		            disposition == PROCESS_AND_DISCARD_COLLECTION) {
		        _collectionManager._discard();
		    }
		
		// otherwise ignore the collection
    		} else {
	    	    _collectionManager._ignore();
    		}
	}
    
  	/** 
    	 * Initializes state of actor at start of workflow execution.
    	 */
	public void initialize() throws IllegalActionException {
 
	    // call superclass constructor
        	super.initialize();
        
        	// clear references to collection manager and parent collection
        	// manager
        	_collectionManager = null;
		_parentCollectionManager = null;
		
		// create a new stack to hold nested collection managers
     	_collectionManagerStack = new Stack();
     	
     	// we start below the upper collections to ignore if  the 
     	// ignoreCollectionsAboveType parameter is the empty string; 
     	// otherwise we are above it
		_belowUpperIgnoredTypes = 
		    _collectionPath.ignoreCollectionsOutside() == Object.class;
		_highestProcessedCollection = null;

		// we always start above the lowest collections to ignore at 
		// start of workflow
     	_aboveLowerIgnoredTypes = true;
     	_lowestProcessedCollection = null;
     	
     	// iterate over previously overriden parameters
     	for (Iterator i = _overriddenParametersMap.keySet().iterator(); 
     			i.hasNext(); ) {
     	    
     	    // change the parameter values back to their defaults
     	    Parameter parameter = (Parameter)i.next();
     	    Token value = parameter.getToken();
     	    _handleParameterChange(parameter, value);
     	}
     	
     	// clear table of overridden parameter values
     	_overriddenParametersMap.clear();
    }
   	
    public IncomingCollectionManager manageIncomingCollection(
    		Collection collection, 
    		CollectionManager parentCollectionManager, 
    		CollectionPort outputPort) 
    		throws IllegalActionException {
	
		return new IncomingCollectionManager(
	        collection, parentCollectionManager, outputPort);
	}
   	    
	public CollectionManager manageNewCollection(
        	Collection collection, CollectionManager parentCollectionManager)
		throws IllegalActionException {

		return new NewCollectionManager(collection, 
		        parentCollectionManager, parentCollectionManager.outputPort());
	}	
	
	public CollectionManager manageNewCollection(
        Collection collection, CollectionManager parentCollectionManager, 
        CollectionPort outputPort) throws IllegalActionException {

		return new NewCollectionManager(
		        collection, parentCollectionManager, outputPort);
	}
	
	/** 
     * Adds a port to the actor at runtime.  This method is called by 
     * Vergil when the user interactively adds ports to actors from within 
     * Vergil and when an actor containing such a port is initialized from
     * its MOML description. This method should not be called under any 
     * other circumstances.
     * 
     * @param name The name of the port to create.
     * @return The new port.
     */ 
    	public Port newPort(String name) 
    		throws NameDuplicationException {
    	    
    		try {
    		    // get a write mutex for the workspace
            	_workspace.getWriteAccess();
            	
            	// construct the new port
            	CollectionIOPort port = CollectionIOPort.createInputPort(this, name);
            	
            	// return the port
            	return port;
            	
        	} catch (IllegalActionException ex) {
        	    
            	// translate to a runtime exception since this should never 
        	    // happen.
            	throw new InternalErrorException(this, ex, null);
            	
        	} finally {
        	    
        	    	// release the workspace mutex
            	_workspace.doneWriting();
        	}
    	}
    	
	///////////////////////////////////////////////////////////////////
	////                         protected methods                 ////

	protected final void _enableParameterOverride(Parameter parameter)
		throws IllegalActionException {
	    
	    if (parameter == null) {
	        throw new NullPointerException(
                "Null parameter not allowed.");
	    }

	    if (parameter.getContainer() != this) {
	        throw new IllegalActionException(
                "Parameter '" + parameter.getName() + 
                "' is not contained by actor " + this.getName() + ".");
	    }
	    
   	    	_overridableParametersMap.put(parameter.getName(), parameter);
	}

	protected final void _enableParameterOverride(Parameter[] parameters) 
		throws IllegalActionException {
	    
	    if (parameters == null) {
	        throw new NullPointerException(
                "Null parameter array not allowed.");
	    }

	    for (int i = 0; i < parameters.length; i++) {
	        _enableParameterOverride(parameters[i]);
	    }
	}

	protected CollectionIOPort _getDefaultOutputPort() {
	    return null;
	}
	
	protected final CollectionIOPort _getOutputPortNamed(String portName) {
	    
	    if (portName == null) {
	        throw new NullPointerException("Null portName not allowed.");
	    }
	    
	    CollectionIOPort port = null;
	    
	    for (Iterator i = outputPortList().iterator(); i.hasNext(); ) {
	        
	        IOPort nextPort = (IOPort)i.next();
	        
	        if (nextPort.getName().equals(portName)) {
	            
	            port = (CollectionIOPort)nextPort;
	            break;
	        }
	    }
	    
	    return port;
	}

	protected void _handleCollectionEnd(CollectionManager collectionManager) 
		throws IllegalActionException, CollectionException {
    }

	protected ExceptionDisposition handleCollectionException(CollectionManager
        collectionManager, CollectionException exception) {

		return PROPAGATE_EXCEPTION;
	}
	
	private void processCollectionException(CollectionException exception) {

	    ExceptionDisposition exceptionDisposition = 
	        handleCollectionException(_collectionManager, exception);
        
	    if (exceptionDisposition == PROPAGATE_EXCEPTION) {
		    _collectionManager.addException(this, exception);
	        ((IncomingCollectionManager)_collectionManager)._ignore();
	    }
    }

    /** 
    	 * Default collection-start handler for collection actors. This
    	 * implementation simply returns 
    	 * {@link #PROCESS_AND_FORWARD_COLLECTION}. Sub-classes may override
    	 * this method to do actor-specific tasks when a new collection
    	 * arrives.  Sub-classes should use the return value to specify 
    	 * what should be done with the collection.  Subclasses may return
    	 * one of four values:
    	 * <p>
    	 * <ul>
    	 * <li>{@link #PROCESS_AND_FORWARD_COLLECTION}
    	 * <li>{@link #PROCESS_AND_DISCARD_COLLECTION}
    	 * <li>{@link #IGNORE_AND_FORWARD_COLLECTION}
    	 * <li>{@link #IGNORE_AND_DISCARD_COLLECTION}
    	 * </ul>
    	 * The <code>handleData</code>, <code>handleMetadata</code>, and
    	 * <code>handleCollectionEnd</code> methods will be called 
    	 * for each data token, each metadata token, and the end delimiter, 
    	 * respectively, received by the actor if the return values indicates 
    	 * that the collection is processed (not ignored). The collection will
    	 * be passed through to succeeding actors in the workflow if the
    	 * collection is forwarded (not discarded).
    	 * 
    	 * @param collectionManager A reference to the 
    	 * {@link org.nddp.CollectionManager} object associated 
    	 * with the incoming collection.
    	 * @return A {@link CollectionDisposition} value of 
    	 * {@link #PROCESS_AND_FORWARD_COLLECTION}. 
    	 */
   	protected CollectionDisposition _handleCollectionStart(CollectionManager 
        collectionManager) throws IllegalActionException, CollectionException {
   	    
   		return PROCESS_AND_FORWARD_COLLECTION;
   	}
   	
	/** 
    	 * Default data token handler for collection actors. This 
    	 * implementation simply returns  {@link #FORWARD_TOKEN}.  Sub-classes
    	 * may override this method in order to receive each data token as it
    	 * arrives.  Sub-classes should use the return value to specify what
    	 * should be done with the token.  Subclasses may return one of two 
    	 * values:
    	 * <p>
    	 * <ul>
    	 * <li>{@link #FORWARD_TOKEN}
    	 * <li>{@link #DISCARD_TOKEN}
    	 * </ul>
    	 * The data will be forwarded on to the succeeding actors in the
    	 * workflow if {@link #FORWARD_TOKEN} is returned and the return value
    	 * from {@link #handleCollectionStart} was 
    	 * {@link #PROCESS_AND_FORWARD_COLLECTION}.  Otherwise, the data token
    	 * is discarded.
    	 * 
    	 * @param collectionManager A reference to the 
    	 * {@link org.nddp.CollectionManager} object associated 
    	 * with the collection containing the data token.
    	 * @param token The data token.
    	 * @return A {@link CollectionDisposition} value of 
    	 * {@link #PROCESS_AND_FORWARD_COLLECTION}. 
	 * @throws CollectionException
	 * @throws 
    	 */
   	protected TokenDisposition _handleData(CollectionManager collectionManager, 
        Token token) throws IllegalActionException, CollectionException {
 		
   	        return _defaultDataDisposition;
  	}
   	
   	protected TokenDisposition _handleDomainObject(CollectionManager 
        collectionManager, DomainObject object) throws IllegalActionException,
        CollectionException {
   	    
   	    return _defaultDataDisposition;
   	}
   	
	protected TokenDisposition _handleExceptionToken(CollectionManager
        collectionManager, ExceptionToken token) throws IllegalActionException,
        CollectionException  {
	    
	    return FORWARD_TOKEN;
    }

	public void handleLoopTerminationToken(LoopTerminationToken
        token) throws IllegalActionException {
        
        throw new IllegalActionException("Actor " + this + 
                " does not support loop termination tokens");
    }       
    
	protected TokenDisposition _handleMetadata(CollectionManager collectionManager, 
        MetadataToken token) throws IllegalActionException, 
        CollectionException {
  	    
  	    return _defaultMetadataDisposition;
    }

  	protected void _handleParameterChange(Parameter parameter, Token value) 
		throws IllegalActionException {
  	    
		super.attributeChanged(parameter);
    }

	protected final CollectionPort _inputPort() {
	    return _inputPort;
	}
    
   	protected final void _setDefaultDataDisposition(TokenDisposition disposition) {
   	    _defaultDataDisposition = disposition;
   	}

   	protected final void _setDefaultMetadataDisposition(TokenDisposition disposition) {
   	    _defaultMetadataDisposition = disposition;
   	}
  	
   	/** 
    	 * Sets the input port used for receiving collections.  This method is
    	 * called by the {@link org.nddp.actors.CollectionSource} class.  
    	 * Subclasses of this class need not call this function.
   	 * 
 	 * @param port The name of the port to use for receiving collections.
    	 */
    protected final void _setInputPort(CollectionPort port) {
	    
	    // set the input port for receiving collections to the passed value
		_inputPort = port;
	}
    
	///////////////////////////////////////////////////////////////////
	////                         private methods                   ////
	
   	private final void _handleVariableToken(VariableToken variableToken) 
		throws IllegalActionException {
   	    	 	    
 	    String parameterName = variableToken.key();
 	    Parameter parameter = 
 	        (Parameter)_overridableParametersMap.get(parameterName);
 	    
 	    if (parameter != null) {
 	   		
 	        Token previousValue;
 	        if (_overriddenParametersMap.containsKey(parameter)) {
 	            previousValue = (Token)_overriddenParametersMap.get(parameter);
 	        } else {
	 	        // previous value is the token stored in the parameter if 
 	            // this is the first override of the parameter
	 	        previousValue = parameter.getToken();
 	        }
 	        
 	        Token newValue = (Token)variableToken.value();
 	        _collectionManager._overrideParameter(parameter, previousValue);
 	        _overriddenParametersMap.put(parameter, newValue);
 	        _handleParameterChange(parameter, newValue);
 	    }    
   	}
   	
   	private boolean _typeMatchesTarget(Token token) {
   	    
   	    if (token instanceof DataToken) {
   	        
            _unwrappedDataToken = ((DataToken)token).getValue();
            
	   	    return _collectionPath.targetDataType().
	   	    		isInstance(_unwrappedDataToken);
   	    
   	    } else if (token instanceof DomainObjectToken) {
   	        
   	        _unwrappedDomainObject = ((DomainObjectToken)token).object();
   	        
			return _collectionPath.targetDataType().
				isInstance(_unwrappedDomainObject);
			
   	    } else {
   	        
   	        _unwrappedDataToken = token;
   	        
	   	    return _collectionPath.targetDataType().
	   	    		isInstance(_unwrappedDataToken);
   	    }
   	}
   	
   	TestProbe createTestProbe() {
   	    return new TestProbe();
   	}

	///////////////////////////////////////////////////////////////////
	////                         private variables                 ////
	
    	private boolean _belowUpperIgnoredTypes;
    	
   	private boolean _aboveLowerIgnoredTypes;
   	
   	private CollectionPath _collectionPath;

    	private IncomingCollectionManager _collectionManager;
    	
	private Stack _collectionManagerStack;
	
	private TokenDisposition _defaultDataDisposition = FORWARD_TOKEN;

	private TokenDisposition _defaultMetadataDisposition = FORWARD_TOKEN;

	private Collection _highestProcessedCollection;
    	
    	private CollectionPort _inputPort;

   	private Collection _lowestProcessedCollection;
    	    	
	private Map _overridableParametersMap = new HashMap();
	
	private Map _overriddenParametersMap = new HashMap();	

	private IncomingCollectionManager _parentCollectionManager;

	private Token _unwrappedDataToken;

	private DomainObject _unwrappedDomainObject;
	
	///////////////////////////////////////////////////////////////////
	////                       public inner classes                ////
   	
	public static class CollectionDisposition {
	    	private CollectionDisposition() {}
	}
	
	public static class ExceptionDisposition {
	    	private ExceptionDisposition() {}
	}
	
    	public static class TokenDisposition {
		private TokenDisposition() {}
    }
    	
    	class TestProbe {
	
    	    public CollectionManager collectionManager() {
    	        return _collectionManager;
    	    }
    	    
		public Map overridableParametersMap() {
		    return _overridableParametersMap;
		}
	
    	}
}
