/* JUnit tests for CollectionPath class.
 *
 * Copyright (c) 2005 Natural Diversity Discovery Project.
 * All rights reserved.
 * 
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 * 
 * IN NO EVENT SHALL THE NATURAL DIVERSITY DISCOVERY PROJECT BE LIABLE 
 * TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR 
 * CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF THIS SOFTWARE AND 
 * ITS DOCUMENTATION, EVEN IF THE NATURAL DIVERSITY DISCOVERY PROJECT
 * HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE NATURAL DIVERSITY DISCOVERY PROJECT SPECIFICALLY DISCLAIMS ANY 
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE NATURAL 
 * DIVERSITY DISCOVERY PROJECT HAS NO OBLIGATION TO PROVIDE MAINTENANCE, 
 * SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.

   @ProposedRating Red (tmcphillips@naturaldiversity.org)
   @AcceptedRating Red (tmcphillips@naturaldiversity.org) 
 */
 
package org.nddp;

import org.nddp.CollectionPath;
import org.nddp.CollectionTypes;
import org.nddp.DataTypes;
import org.nddp.DataTypes.NoMatch;

import ptolemy.kernel.util.IllegalActionException;
import junit.framework.TestCase;

public class TestCollectionPath extends TestCase {

	///////////////////////////////////////////////////////////////////
	////                         public methods                    //// 
    
    public void setUp() throws Exception {

        	CollectionTypes.clear();
		CollectionTypes.register("C", C.class);
		CollectionTypes.register("E", E.class);

		DataTypes.register("D", D.class);
    }

    public void testConstructor_BadPathFormatException() throws Exception {
        
        _verifyBadPathException("/");
        _verifyBadPathException("/C");
        _verifyBadPathException("C/C/C");
        _verifyBadPathException("//C/");
        _verifyBadPathException("///C");
    }
    
    public void testConstructor_Collection() throws Exception {

		CollectionPath path = new CollectionPath("C");
		_verifyState(path, C.class, C.class, NoMatch.class, true);
    }
    
    public void testConstructor_CollectionSlash() throws Exception {

		CollectionPath path = new CollectionPath("C/");
		_verifyState(path, C.class, C.class, Object.class, true);
    }
    
    public void testConstructor_CollectionSlashData() throws Exception {

		CollectionPath path = new CollectionPath("C/D");
		_verifyState(path, C.class, C.class, D.class, true);
    }
    
    public void testConstructor_CollectionSlashSlash() throws Exception {

		CollectionPath path = new CollectionPath("C//");
		_verifyState(path, C.class, Object.class, Object.class, true);
    }

    public void testConstructor_CollectionSlashSlashCollection() 
    		throws Exception {

		CollectionPath path = new CollectionPath("C//E");
		_verifyState(path, C.class, E.class, Object.class, true);
    }
    
    public void testConstructor_CollectionSlashSlashData() throws Exception {

		CollectionPath path = new CollectionPath("C//D");
		_verifyState(path, C.class, Object.class, D.class, true);
    }
    
    public void testConstructor_CollectionSlashSlashDataSlashSlashCollection()
    		throws Exception {

		CollectionPath path = new CollectionPath("C//D//E");
		_verifyState(path, C.class, E.class, D.class, true);
    }
    
    public void testConstructor_Data() throws Exception {

		CollectionPath path = new CollectionPath("D");
		_verifyState(path, Object.class, Object.class, D.class, false);
    }
    
    public void testConstructor_DataSlashSlashCollection() throws Exception {

		CollectionPath path = new CollectionPath("D//C");
		_verifyState(path, Object.class, C.class, D.class, false);
    }
    
    public void testConstructor_EmptyPath() throws Exception {
        
        CollectionPath path = new CollectionPath("");
		_verifyState(path, Object.class, Object.class, Object.class, true);
    }
    
    public void testConstructor_SlashSlashCollection() throws Exception {

		CollectionPath path = new CollectionPath("//C");
		_verifyState(path, Object.class, C.class, Object.class, true);
    }

    public void testConstructor_SlashSlashData() throws Exception {

		CollectionPath path = new CollectionPath("//D");
		_verifyState(path, Object.class, Object.class, D.class, true);
    }
    
    public void testConstructor_SlashSlashDataSlashSlashCollection() 
    		throws Exception {

		CollectionPath path = new CollectionPath("//D//C");
		_verifyState(path, Object.class, C.class, D.class, true);
    }
    
    public void testConstructor_UndefinedCollectionException() 
    		throws Exception {
        
        _verifyUndefinedCollectionException("F/", "F");
        _verifyUndefinedCollectionException("F/D", "F");
        _verifyUndefinedCollectionException("F//", "F");
        _verifyUndefinedCollectionException("F//D//C", "F");
        _verifyUndefinedCollectionException("C//D//F", "F");
    }

    public void testConstructor_UndefinedCollectionOrDataException() 
    		throws Exception {
        
        _verifyUndefinedCollectionOrDataException("F", "F");
        _verifyUndefinedCollectionOrDataException("C//F", "F");
        _verifyUndefinedCollectionOrDataException("F//D", "F");
        _verifyUndefinedCollectionOrDataException("//F", "F");
        _verifyUndefinedCollectionOrDataException("F//C", "F");
    }

    public void testConstructor_WhiteSpaceException() throws Exception {
        
        _verifyWhiteSpaceException("C /");
        _verifyWhiteSpaceException("C /D");
        _verifyWhiteSpaceException("C/ D");
        _verifyWhiteSpaceException("C //");
        _verifyWhiteSpaceException("C/ /");
        _verifyWhiteSpaceException("C// D");
        _verifyWhiteSpaceException("C/ /D");
        _verifyWhiteSpaceException("C //D");
        _verifyWhiteSpaceException("// C");
        _verifyWhiteSpaceException("/ /C");
        _verifyWhiteSpaceException("C //D//E");
        _verifyWhiteSpaceException("C/ /D//E");
        _verifyWhiteSpaceException("C// D//E");
        _verifyWhiteSpaceException("C//D //E");
        _verifyWhiteSpaceException("C//D/ /E");
        _verifyWhiteSpaceException("C//D// E");
    }

	///////////////////////////////////////////////////////////////////
	////                         private methods                   ////

    private void _verifyBadPathException(String pathString) throws Exception {
        
        try {
	        CollectionPath path = new CollectionPath(pathString);
        		fail("An IllegalActionException should have been thrown.");
        }
        catch(IllegalActionException expected) {
            assertEquals("Path " + pathString + 
                " does not match expected format.", expected.getMessage());
        }    
    }
        
    private void _verifyState(CollectionPath path, Class outside, 
        Class inside, Class type, boolean expose) {
        
        assertEquals(outside, path.ignoreCollectionsOutside()); 
        assertEquals(inside, path.ignoreCollectionsInside());     
        assertEquals(type, path.targetDataType());
        assertEquals(expose, path.exposeNonDataEvents());
    }    
        
    private void _verifyUndefinedCollectionException(String pathString, 
        String collectionTypeName) throws Exception {
        
        try {
	        CollectionPath path = new CollectionPath(pathString);
        		fail("An IllegalActionException should have been thrown.");
        }
        catch(IllegalActionException expected) {
            assertEquals("No collection type " + collectionTypeName, 
                expected.getMessage());
        }    
    }
    
    private void _verifyUndefinedCollectionOrDataException(String pathString, 
        String collectionTypeName) throws Exception {
        
        try {
	        CollectionPath path = new CollectionPath(pathString);
        		fail("An IllegalActionException should have been thrown.");
        }
        catch(IllegalActionException expected) {
            assertEquals("No collection or data type " + collectionTypeName +
                ".", expected.getMessage());
        }    
    }
    
    private void _verifyWhiteSpaceException(String pathString) 
    		throws Exception {
        
        try {
	        CollectionPath path = new CollectionPath(pathString);
        		fail("An IllegalActionException should have been thrown.");
        }
        catch(IllegalActionException expected) {
            assertEquals("No white space is allowed in collection path: " + 
                pathString, expected.getMessage());
        }    
    }

    	///////////////////////////////////////////////////////////////////
	////                   private innter classes                  ////   
	    
	class C extends Collection {
	
	    public C(String name) {
	        super(name);
	    }
	    
	    public String type() {
	        return "C";
	    }
	}
	
	class D {}
	
	class E extends Collection {
	
	    public E(String name) {
	        super(name);
	    }
	    
	    public String type() {
	        return "E";
	    }
	}
}

