/* JUnit test class for DataType class.
 *
 * Copyright (c) 2005 Natural Diversity Discovery Project.
 * All rights reserved.
 * 
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 * 
 * IN NO EVENT SHALL THE NATURAL DIVERSITY DISCOVERY PROJECT BE LIABLE 
 * TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR 
 * CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF THIS SOFTWARE AND 
 * ITS DOCUMENTATION, EVEN IF THE NATURAL DIVERSITY DISCOVERY PROJECT
 * HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE NATURAL DIVERSITY DISCOVERY PROJECT SPECIFICALLY DISCLAIMS ANY 
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE NATURAL 
 * DIVERSITY DISCOVERY PROJECT HAS NO OBLIGATION TO PROVIDE MAINTENANCE, 
 * SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.

   @ProposedRating Red (tmcphillips@naturaldiversity.org)
   @AcceptedRating Red (tmcphillips@naturaldiversity.org) 
 */
 
package org.nddp;

import org.nddp.DataTypes;
import org.nddp.exceptions.CollectionException;
import org.nddp.util.Xml.Indentation;

import ptolemy.data.DoubleToken;
import ptolemy.data.IntToken;
import ptolemy.data.ScalarToken;
import ptolemy.data.StringToken;
import ptolemy.data.expr.Parameter;
import ptolemy.kernel.util.IllegalActionException;
import ptolemy.kernel.util.NamedObj;
import junit.framework.TestCase;

public class TestDataTypes extends TestCase {

    public TestDataTypes() {
        super();
    }
   
	///////////////////////////////////////////////////////////////////
	////                         public methods                    //// 
 
    public void setUp() {
        	DataTypes.clear();
    }
    
    public void testClear() throws Exception {
        
		DataTypes.register("A", ClassA.class);
        	DataTypes.clear();

        	try {
   	        	DataTypes.valueOfString("A");

   	        	fail("An IllegalActionException should be thrown because _clear() " +
   	        	        "was called after adding ClassA.");
        	} 
        catch(IllegalActionException expected) {
            assertEquals("No data type A", expected.getMessage());
        }
    }
    
    public void testGetDoubleProperty_PublicDoubleProperty() throws Exception {
        
        DomainObject b = new DomainObjectB();
        
        assertEquals(3.14159, 
                DataTypes.getDoubleProperty(b, "publicDoubleProperty"), 0.00001);
    }
     
   
    	public void testGetDoubleProperty_PublicDoubleStringProperty() throws Exception {
        
        DomainObject b = new DomainObjectB();
        
        assertEquals(2.71828183, 
                DataTypes.getDoubleProperty(b, "publicDoubleStringProperty"), 0.00000001);
    }
    	
    	public void testGetDoubleProperty_PublicIntProperty() throws Exception {
        
        DomainObject b = new DomainObjectB();
        
        assertEquals(42, 
                DataTypes.getDoubleProperty(b, "publicIntProperty"), 0.00001);
    	}
    	
    	public void testGetDoubleProperty_PublicStringProperty() throws Exception {
        
        	DomainObject b = new DomainObjectB();
		
        	try {
        		DataTypes.getDoubleProperty(b, "publicStringProperty");
	
	        	fail("A NumberFormatException should have been thrown.");
        	} 
       	catch(NumberFormatException expected) {
        	}
    	}    	
    	
    public void testGetProperty_MissingProperty() throws Exception {
        
        DomainObject b = new DomainObjectB();
        
        try {
	        DataTypes.getProperty(b, "missingProperty");
	        
	        fail("An IllegalActionException should have been thrown.");
        }
        catch(IllegalActionException expected) {
            assertEquals("Unable to access missingProperty() method for class"
            		+ " org.nddp.TestDataTypes$DomainObjectB", expected.getMessage());
        }
    }
    
    public void testGetProperty_ProtectedProperty() throws Exception {
        
        DomainObject b = new DomainObjectB();
        
        try {
	        DataTypes.getProperty(b, "protectedProperty");
	        
	        fail("An IllegalActionException should have been thrown.");
        }
        catch(IllegalActionException expected) {
            assertEquals("Unable to access protectedProperty() method for class"
            		+ " org.nddp.TestDataTypes$DomainObjectB", expected.getMessage());
        }
    	}    

    	public void testGetProperty_PublicVoidProperty() throws Exception {
        
        DomainObject b = new DomainObjectB();
        
        try {
	        DataTypes.getProperty(b, "publicVoidProperty");
	        
	        fail("An NullPointerException should have been thrown.");
        }
        catch(NullPointerException expected) {
            assertEquals("Null returned for property publicVoidProperty for class"
            		+ " org.nddp.TestDataTypes$DomainObjectB", expected.getMessage());
        }
    }
    	
    public void testGetStringProperty_PublicBooleanProperty() throws Exception {
        
        DomainObject b = new DomainObjectB();
        
        assertEquals("true", 
                DataTypes.getStringProperty(b, "publicBooleanProperty"));
    } 
    
    public void testGetStringProperty_PublicDoubleProperty() throws Exception {
        
        DomainObject b = new DomainObjectB();
        
        assertEquals("3.14159", 
                DataTypes.getStringProperty(b, "publicDoubleProperty"));
    }
    
    public void testGetStringProperty_PublicIntProperty() throws Exception {
        
        DomainObject b = new DomainObjectB();
        
        assertEquals("42", 
                DataTypes.getStringProperty(b, "publicIntProperty"));
    }    	
    
    public void testGetStringProperty_PublicStringProperty() throws Exception {
        
        DomainObject b = new DomainObjectB();
        
        assertEquals("public string value", 
                DataTypes.getStringProperty(b, "publicStringProperty"));
    }
    
    public void testIsDefined_EmptyTypeName() throws Exception {

		assertEquals(false, DataTypes.isDefined(""));
    }
    
    public void testIsDefined_Match() throws Exception {

        	DataTypes.register("A", ClassA.class);
		assertEquals(true, DataTypes.isDefined("A"));
    }

    public void testIsDefined_NoMatch() throws Exception {

		assertEquals(false, DataTypes.isDefined("A"));
    }
    
    public void testIsDefined_NullTypeName() throws Exception {

        try {
			DataTypes.isDefined(null);
			
			fail("A NullPointerException should have been thrown.");
        }
        catch(NullPointerException expected) {
            assertEquals("Null typeName not allowed.", 
            		expected.getMessage());
        }
    }    

    public void testRegister() throws Exception {

		DataTypes.register("A", ClassA.class);
		assertEquals(ClassA.class, DataTypes.valueOfString("A"));
	}
    
    public void testRegister_NoMatchClass() throws Exception {

        try  {
			DataTypes.register("B", DataTypes.NoMatch.class);
			
			fail("An AssertionError should be thrown because NoMatch " +
		        "was passed for typeClass");
        }
        catch(AssertionError expected) {
            assertEquals("Class NoMatch not allowed", 
            		expected.getMessage());
        }
	}
    
    public void testRegister_NullTypeClass() throws Exception {

        try  {
			DataTypes.register("B", null);
			
			fail("An AssertionError should be thrown because null " +
		        "was passed for typeClass");
        }
        catch(AssertionError expected) {
            assertEquals("Null typeClass not allowed.", 
            		expected.getMessage());
        }
	}
    
    public void testRegister_NullTypeName() throws Exception {

        try  {
			DataTypes.register(null, ClassA.class);
			
			fail("An AssertionError should be thrown because null" +
		        "was passed for typeName");
        }
        catch(AssertionError expected) {
            assertEquals("Null typeName not allowed.", 
            		expected.getMessage());
        }
	}

    public void testRegisterDefaultDataTypes() throws Exception {
        
        DataTypes.registerDefaultDataTypes();
        
        assertEquals(DoubleToken.class, DataTypes.valueOfString("Double"));
        assertEquals(IntToken.class, DataTypes.valueOfString("Integer"));
        assertEquals(ScalarToken.class, DataTypes.valueOfString("Scalar"));
        assertEquals(StringToken.class, DataTypes.valueOfString("String"));        
    }
        
    public void testValueOfParameter_EmptyClassName() throws Exception {
        
        	DataTypes.registerDefaultDataTypes();
        	Parameter parameter = new Parameter(
			new NamedObj(), "parameter", new StringToken(""));
        	assertEquals(Object.class, DataTypes.valueOfParameter(parameter));
    } 
    
    public void testValueOfParameter_Match() throws Exception {
        
		DataTypes.register("A", ClassA.class);
		
        	Parameter parameter = new Parameter(
			new NamedObj(), "parameter", new StringToken("A"));
        	
        	assertEquals(ClassA.class, DataTypes.valueOfParameter(parameter));
    }
    
    public void testValueOfParameter_NoMatch() throws Exception {
        
        	Parameter parameter = new Parameter(
			new NamedObj(), "parameter", new StringToken("A"));
        	
        	try { 
        	    DataTypes.valueOfParameter(parameter);
	        
	        fail("An IllegalActionException should be thrown" +
	                "because ClassA has not been added.");
        	}
         catch(IllegalActionException expected) {
            assertEquals("No data type A", expected.getMessage());
        }
    }
    
    public void testValueOfParameter_NullParameter() throws Exception {
        
        	try {
	        DataTypes.valueOfParameter(null);
	        
	        fail("A NullPointerException should be thrown" +
	        		"because null was passed for the parameter.");
        	}
        catch(NullPointerException expected) {
            assertEquals("Null dataTypeParameter not allowed.", 
            		expected.getMessage());
        }
    }

    public void testValueOfString_EmptyClassName() throws Exception {
        
        DataTypes.registerDefaultDataTypes();
        assertEquals(Object.class, DataTypes.valueOfString(""));
    }
    
    public void testValueOfString_Match() throws Exception {
        
		DataTypes.register("A", ClassA.class);
        	assertEquals(ClassA.class, DataTypes.valueOfString("A"));
    }
    
    public void testValueOfString_NoMatch() throws Exception {
        
        try {
	        DataTypes.valueOfString("A");
	        
	        fail("An IllegalActionException should be thrown" +
	                "because ClassA has not been added.");
        }
        catch(IllegalActionException expected) {
            assertEquals("No data type A", expected.getMessage());
        }
    }

    public void testValueOfString_NullTypeName() throws Exception {
        
        	try {
	        DataTypes.valueOfString(null);
	        
	        fail("A NullPointerException should have been thrown.");
        	}
        catch(NullPointerException expected) {
            assertEquals("Null typeName not allowed.", 
            		expected.getMessage());
        }
    }

    ///////////////////////////////////////////////////////////////////
	////                  private inner classes                    //// 
    
	private class ClassA {}
	
	private class DomainObjectB implements DomainObject {
	
		///////////////////////////////////////////////////////////////////
		////                         public methods                    //// 
	    
	    public String id() {
	        return null;
	    }
	    
	    public boolean publicBooleanProperty() {    
	        return true;
	    }
	    
	    public double publicDoubleProperty() {    
	        return 3.14159;
	    }
	    
	    public String publicDoubleStringProperty() {
	        return "2.71828183";
	    }
	
	    public int publicIntProperty() {    
	        return 42;
	    }
	    
	    public String publicStringProperty() {
	        return "public string value";
	    }
	    
	    public void publicVoidProperty() {
	    }
	
	    public String xmlAttributeString() {
	        return null;
	    }
	
	    public String xmlContentString(Indentation indentation) {
	        return null;
	    }
	    
		///////////////////////////////////////////////////////////////////
		////                        private methods                    //// 
	
	    private String privateStringProperty() {
	        return "private string value";
	    }
	    
	    private String protectedStringProperty() {
	        return "protected string value";
	    }

        public void setWriteLock() {
        }

        public void assertNotWriteLocked() throws CollectionException {
        }
	}
}	
